/*
 * fvkbd-keyboard.c keyboard unit for fvkbd  
 *
 * Copyright (C) 2009, Intel Corporation.
 *
 * Author: Raymond Liu <raymond.liu@intel.com>
 * 
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 * 
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */


#include "fvkbd.h"

#include "fvkbd-keyboard.h"
#include "fvkbd-panel.h"
#include "misc-utility.h"


#define FVKBD_KEYBOARD_GET_PRIVATE(object)\
	(G_TYPE_INSTANCE_GET_PRIVATE((object), FVKBD_TYPE_KEYBOARD, FvkbdKeyboardPrivate))

struct _FvkbdKeyboardPrivate {
	gint modes_num;
	gint current_mode;
	gboolean in_temp_mode;
	GSList *modes;
	GSList *panels;
	FakeKey *fakekey;

	gchar *font_family;
	gchar *font_weight;
	gint font_size;

	gchar *efont_family;
	gchar *efont_weight;
	gint efont_size;

	gchar *ofont_family;
	gchar *ofont_weight;
	gint ofont_size;
};

G_DEFINE_TYPE (FvkbdKeyboard, fvkbd_keyboard, FVKBD_TYPE_UNIT)

typedef struct _FvkbdKeyboardMode FvkbdKeyboardMode;
struct _FvkbdKeyboardMode {
	gint id;
	gchar *name;
};

FvkbdKeyboard *keyboard = NULL;
FakeKey *fakekey_instance = NULL;

static gboolean fvkbd_keyboard_parse_xml (FvkbdUnit *unit, FvkbdParser *parser);


gint
fvkbd_keyboard_get_modes_number (FvkbdKeyboard *self)
{
	FvkbdKeyboardPrivate *priv;

	g_return_val_if_fail(FVKBD_IS_KEYBOARD(self), -1);

	priv = FVKBD_KEYBOARD_GET_PRIVATE(self);
	return priv->modes_num;
}


GSList *
fvkbd_keyboard_get_panels (FvkbdKeyboard *self)
{
	FvkbdKeyboardPrivate *priv;

	g_return_val_if_fail(FVKBD_IS_KEYBOARD(self), NULL);

	priv = FVKBD_KEYBOARD_GET_PRIVATE(self);
	return priv->panels;
}


gint
fvkbd_keyboard_get_current_mode (void)
{
	FvkbdKeyboardPrivate *priv;

	if (keyboard == NULL)
		return -1;

	priv = FVKBD_KEYBOARD_GET_PRIVATE(keyboard);

	return priv->current_mode;
}


inline gboolean
fvkbd_in_temp_mode(void)
{
	FvkbdKeyboardPrivate *priv;

	if (keyboard == NULL)
		return FALSE;

	priv = FVKBD_KEYBOARD_GET_PRIVATE(keyboard);

	return priv->in_temp_mode;
}


FakeKey *
fvkbd_keyboard_get_fakekey (void)
{
	FvkbdKeyboardPrivate *priv;

	if (keyboard == NULL)
		return NULL;

	priv = FVKBD_KEYBOARD_GET_PRIVATE(keyboard);

	return priv->fakekey;
}


gchar*
fvkbd_keyboard_get_font_family (KbdFontType type)
{
	FvkbdKeyboardPrivate *priv;

	if (keyboard == NULL)
		return NULL;

	priv = FVKBD_KEYBOARD_GET_PRIVATE(keyboard);

	switch (type) {
		case KBD_FONT_TYPE_NORMAL:
			return priv->font_family;
		case KBD_FONT_TYPE_EXTRA:
			return priv->efont_family;
		case KBD_FONT_TYPE_POP:
			return priv->ofont_family;
		default:
			return NULL;
	}

	return NULL;
}


gchar*
fvkbd_keyboard_get_font_weight (KbdFontType type)
{
	FvkbdKeyboardPrivate *priv;

	if (keyboard == NULL)
		return NULL;

	priv = FVKBD_KEYBOARD_GET_PRIVATE(keyboard);

	switch (type) {
		case KBD_FONT_TYPE_NORMAL:
			return priv->font_weight;
		case KBD_FONT_TYPE_EXTRA:
			return priv->efont_weight;
		case KBD_FONT_TYPE_POP:
			return priv->ofont_weight;
		default:
			return NULL;
	}

	return NULL;
}


gint
fvkbd_keyboard_get_font_size (KbdFontType type)
{
	FvkbdKeyboardPrivate *priv;

	if (keyboard == NULL)
		return -1;

	priv = FVKBD_KEYBOARD_GET_PRIVATE(keyboard);

	switch (type) {
		case KBD_FONT_TYPE_NORMAL:
			return priv->font_size;
		case KBD_FONT_TYPE_EXTRA:
			return priv->efont_size;
		case KBD_FONT_TYPE_POP:
			return priv->ofont_size;
		default:
			return 0;
	}

	return 0;
}


gint
fvkbd_keyboard_change_mode (gint id, gboolean lock)
{
	FvkbdKeyboardPrivate *priv;
	FvkbdUnit *panel;
	int i;
	gint ret;
	gint previous_mode;

	if (keyboard == NULL)
		return -1;

	priv = FVKBD_KEYBOARD_GET_PRIVATE(keyboard);
	DBG("current mode = %d, new mode = %d\n", priv->current_mode, id);

	if (id >= priv->modes_num)
		return -1;

	previous_mode = priv->current_mode;
	priv->current_mode = id;
	if (!lock)
		priv->in_temp_mode = TRUE;
	else
		priv->in_temp_mode = FALSE;

	for (i = 0; i < g_slist_length(priv->panels); i++) {
		panel = g_slist_nth_data(priv->panels, i);
		if ( (ret  = fvkbd_unit_set_mode(panel, id)) != 0) {
			g_warning("Failed to switch to mode %d\n", id);
			priv->current_mode = previous_mode;
			return ret;
		}
	}

	return 0;
}


gint
fvkbd_keyboard_resume_default_mode (void)
{
	FvkbdKeyboardPrivate *priv;
	FvkbdUnit *panel;
	gint default_mode = 0;
	int i;
	gint ret;


	if (keyboard == NULL)
		return -1;

	priv = FVKBD_KEYBOARD_GET_PRIVATE(keyboard);

	if (priv->current_mode == default_mode)
		return 0;

	priv->current_mode = default_mode;
	priv->in_temp_mode = FALSE;

	for (i = 0; i < g_slist_length(priv->panels); i++) {
		panel = g_slist_nth_data(priv->panels, i);
		if ( (ret  = fvkbd_unit_set_mode(panel, default_mode)) != 0) {
			g_warning("Failed to switch to mode %d\n", 0);
			return ret;
		}
	}

	return 0;
}


#define MIN_KEYBOARD_MODES 1
#define MAX_KEYBOARD_MODES 16


static gint
keyboard_parse_modes (FvkbdUnit *unit, FvkbdParser *parser, void *data)
{
	FvkbdKeyboard *keyboard = FVKBD_KEYBOARD(unit);
	FvkbdKeyboardPrivate *priv = FVKBD_KEYBOARD_GET_PRIVATE(keyboard);

	FvkbdKeyboardMode *mode = NULL;
	gchar *tmp_str = NULL;
	gint id;
	gint ret = FALSE;

	if ((tmp_str = fvkbd_parser_get_attribute(parser, "id")) == NULL) {
		fvkbd_parser_set_error(parser, "Missing attribute : id");
		goto done;
	}

	id = atoi(tmp_str);
	if (id < 0) {
		fvkbd_parser_set_error(parser, "Invalid id value");
		goto done;
	}

	g_free(tmp_str);
	if ((tmp_str = fvkbd_parser_get_attribute(parser, "name")) == NULL) {
		fvkbd_parser_set_error(parser, "Missing attribure : name");
		goto done;
	}

	mode = g_new0(FvkbdKeyboardMode, 1);
	mode->id = id;
	mode->name = g_strdup(tmp_str);
	priv->modes = g_slist_append(priv->modes, mode);

	ret = TRUE;
done:
	g_free(tmp_str);
	return ret;
}


static gint
keyboard_parse_panel (FvkbdUnit *unit, FvkbdParser *parser, void *data)
{
	FvkbdKeyboard *keyboard = FVKBD_KEYBOARD(unit);
	FvkbdKeyboardPrivate *priv = FVKBD_KEYBOARD_GET_PRIVATE(keyboard);
	FvkbdUnit *panel;
	gint ret = FALSE;

	panel = fvkbd_panel_new();
	fvkbd_unit_set_parent(panel, unit);
	if (fvkbd_unit_parse_xml(panel, parser) != TRUE) {
		g_object_unref(panel);
		goto done;
	}

	priv->panels = g_slist_append(priv->panels, panel);
	ret = TRUE;

done:
	return ret;
}


static UnitParserFunc keyboard_element_parsers_1[] = {
	{"mode", keyboard_parse_modes},
	{NULL}
};


static UnitParserFunc keyboard_element_parsers_2[] = {
	{"panel", keyboard_parse_panel},
	{NULL}
};


static void
fvkbd_keyboard_finalize (GObject *object)
{
	FvkbdKeyboard *kbd = FVKBD_KEYBOARD(object);
	FvkbdKeyboardPrivate *priv = FVKBD_KEYBOARD_GET_PRIVATE(kbd);
	GSList *tmp;
	FvkbdKeyboardMode *mode = NULL;

	for (tmp = priv->modes; tmp; tmp = tmp->next) {
		mode = tmp->data;
		g_free(mode->name);
		g_free(mode);
	}

	g_slist_free(priv->modes);
	priv->modes = NULL;

	for (tmp = priv->panels; tmp; tmp = tmp->next)
		g_object_unref(tmp->data);

	g_slist_free(priv->panels);
	priv->panels = NULL;

	g_free(priv->font_family);
	g_free(priv->efont_family);
	g_free(priv->ofont_family);
	g_free(priv->font_weight);
	g_free(priv->efont_weight);
	g_free(priv->ofont_weight);

	keyboard = NULL;

	G_OBJECT_CLASS(fvkbd_keyboard_parent_class)->finalize(object);
}


static void
fvkbd_keyboard_class_init (FvkbdKeyboardClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
	FvkbdUnitClass *unit_class = FVKBD_UNIT_CLASS(klass);

	g_type_class_add_private(klass, sizeof(FvkbdKeyboardPrivate));

	gobject_class->finalize = fvkbd_keyboard_finalize;

	unit_class->parse_xml = fvkbd_keyboard_parse_xml;
	unit_class->build_ui = fvkbd_keyboard_build_ui;
	unit_class->resize_ui = fvkbd_keyboard_resize_ui;
	unit_class->destroy_ui = fvkbd_keyboard_destroy_ui;
}


static void
fvkbd_keyboard_init (FvkbdKeyboard *self)
{
	FvkbdKeyboardPrivate *priv = FVKBD_KEYBOARD_GET_PRIVATE(self);
	FvkbdUnit *unit = FVKBD_UNIT(self);

	if (fakekey_instance == NULL) {
		Display *xdpy = get_x_display();

		fakekey_instance = fakekey_init(xdpy);
	}
	priv->fakekey = fakekey_instance;
	priv->modes_num = 0;
	priv->current_mode = 0;
	priv->in_temp_mode = FALSE;
	priv->font_family = NULL;
	priv->efont_family = NULL;
	priv->ofont_family = NULL;
	priv->font_weight = NULL;
	priv->efont_weight = NULL;
	priv->ofont_weight = NULL;
	priv->font_size = 0;
	priv->efont_size = 0;
	priv->ofont_size = 0;

	fvkbd_set_unit_type(unit, UNIT_TYPE_KEYBOARD);
}


FvkbdUnit *
fvkbd_keyboard_new (void)
{
	if (keyboard == NULL)
		keyboard = g_object_new(FVKBD_TYPE_KEYBOARD, NULL);

	return FVKBD_UNIT(keyboard);
}


static gboolean fvkbd_keyboard_parse_xml (FvkbdUnit *unit, FvkbdParser *parser)
{
	FvkbdKeyboard *keyboard;
	FvkbdKeyboardPrivate *priv;
	UnitParserFunc *pfp;
	gchar *tmp_str = NULL;
	gint modes_num;
	gint i;
	gint ret = FALSE;

	g_return_val_if_fail(FVKBD_IS_KEYBOARD(unit), -1);
	keyboard = FVKBD_KEYBOARD(unit);
	priv = FVKBD_KEYBOARD_GET_PRIVATE(keyboard);

	// parsing attr num_of modes
	if ((tmp_str = fvkbd_parser_get_attribute(parser, "num_of_modes")) == NULL) {
		fvkbd_parser_set_error(parser, "Missing attribute : num_of_modes");
		goto done;
	}

	modes_num = atoi(tmp_str);
	if ((modes_num < MIN_KEYBOARD_MODES) || (modes_num > MAX_KEYBOARD_MODES)) {
		fvkbd_parser_set_error(parser, "num_of_modes out of valid range");
		goto done;
	}

	priv->modes_num = modes_num;

	// paring attr font, efont, ofont. parse_font_property could deal with NULL
	g_free(tmp_str);
	tmp_str = fvkbd_parser_get_attribute(parser, "font");
	parse_font_property(tmp_str, &priv->font_family, &priv->font_weight, &priv->font_size,
			kbd_get_default_font_family(KBD_FONT_TYPE_NORMAL),
			kbd_get_default_font_weight(KBD_FONT_TYPE_NORMAL),
			kbd_get_default_font_size(KBD_FONT_TYPE_NORMAL));

	g_free(tmp_str);
	tmp_str = fvkbd_parser_get_attribute(parser, "efont");
	parse_font_property(tmp_str, &priv->efont_family, &priv->efont_weight, &priv->efont_size,
			kbd_get_default_font_family(KBD_FONT_TYPE_EXTRA),
			kbd_get_default_font_weight(KBD_FONT_TYPE_EXTRA),
			kbd_get_default_font_size(KBD_FONT_TYPE_EXTRA));

	g_free(tmp_str);
	tmp_str = fvkbd_parser_get_attribute(parser, "ofont");
	parse_font_property(tmp_str, &priv->ofont_family, &priv->ofont_weight, &priv->ofont_size,
			kbd_get_default_font_family(KBD_FONT_TYPE_POP),
			kbd_get_default_font_weight(KBD_FONT_TYPE_POP),
			kbd_get_default_font_size(KBD_FONT_TYPE_POP));

	// parsing color attr
	g_free(tmp_str);
	if ((tmp_str = fvkbd_parser_get_attribute(parser, "panel_bgcolor")) == NULL) {
		fvkbd_unit_set_color(unit, KBD_COLOR_TYPE_PANEL_BG, kbd_get_default_kbd_bg_color());
	} else {
		fvkbd_unit_set_color(unit, KBD_COLOR_TYPE_PANEL_BG, kbd_str_to_color(tmp_str));
	}

	g_free(tmp_str);
	if ((tmp_str = fvkbd_parser_get_attribute(parser, "key_bgcolor")) == NULL) {
		fvkbd_unit_set_color(unit, KBD_COLOR_TYPE_KEY_BG, kbd_get_default_key_bg_color());
	} else {
		fvkbd_unit_set_color(unit, KBD_COLOR_TYPE_KEY_BG, kbd_str_to_color(tmp_str));
	}

	g_free(tmp_str);
	if ((tmp_str = fvkbd_parser_get_attribute(parser, "key_fgcolor")) == NULL) {
		fvkbd_unit_set_color(unit, KBD_COLOR_TYPE_KEY_FG, kbd_get_default_key_fg_color());
	} else {
		fvkbd_unit_set_color(unit, KBD_COLOR_TYPE_KEY_FG, kbd_str_to_color(tmp_str));
	}

	g_free(tmp_str);
	if ((tmp_str = fvkbd_parser_get_attribute(parser, "pop_bgcolor")) == NULL) {
		fvkbd_unit_set_color(unit, KBD_COLOR_TYPE_KEY_POP_BG, kbd_get_default_key_pop_bg_color());
	} else {
		fvkbd_unit_set_color(unit, KBD_COLOR_TYPE_KEY_POP_BG, kbd_str_to_color(tmp_str));
	}

	g_free(tmp_str);
	if ((tmp_str = fvkbd_parser_get_attribute(parser, "pop_fgcolor")) == NULL) {
		fvkbd_unit_set_color(unit, KBD_COLOR_TYPE_KEY_POP_FG, kbd_get_default_key_pop_fg_color());
	} else {
		fvkbd_unit_set_color(unit, KBD_COLOR_TYPE_KEY_POP_FG, kbd_str_to_color(tmp_str));
	}

	g_free(tmp_str);
	if ((tmp_str = fvkbd_parser_get_attribute(parser, "key_extra_fgcolor")) == NULL) {
		fvkbd_unit_set_color(unit, KBD_COLOR_TYPE_KEY_EXTRA_FG,
					kbd_get_default_key_extra_fg_color());
	} else {
		fvkbd_unit_set_color(unit, KBD_COLOR_TYPE_KEY_EXTRA_FG, kbd_str_to_color(tmp_str));
	}

	// parsing children elements
	if (!fvkbd_parser_go_child_element(parser)) {
		fvkbd_parser_set_error(parser, "Missing children");
		goto done;
	}

	// parsing element group 1
	do {
		for (pfp = keyboard_element_parsers_1; pfp->node_name != NULL; pfp++) {
			if (!fvkbd_parser_element_match(parser, pfp->node_name))
				continue;

			if (pfp->parser(unit, parser, NULL) != TRUE)
				goto done;

			break;
		}

		if (pfp->node_name == NULL)
			break;
	} while (fvkbd_parser_element_next(parser));

	if (g_slist_length(priv->modes) != priv->modes_num) {
		fvkbd_parser_set_error(parser, "Mode elements' number not matching");
		goto done;
	}

	for (i = 0; i < g_slist_length(priv->modes); i++) {
		if (((FvkbdKeyboardMode *)g_slist_nth_data(priv->modes, i))->id == 0)
			break;
	}

	if (i == g_slist_length(priv->modes)) {
		fvkbd_parser_set_error(parser, "Could not find default mode with id \"0\"");
		goto done;
	}

	// parsing element group 2
	do {
		for (pfp = keyboard_element_parsers_2; pfp->node_name != NULL; pfp++) {
			if (!fvkbd_parser_element_match(parser, pfp->node_name))
				continue;

			if (pfp->parser(unit, parser, NULL) != TRUE)
				goto done;

			break;
		}
	} while(fvkbd_parser_element_next(parser));

	ret = TRUE;
done:
	g_free(tmp_str);
	fvkbd_parser_go_parent_element(parser);
	return ret;
}

