/*
 * fvkbd-key-ui-gtk.c key unit gtk ui for fvkbd  
 *
 * Copyright (C) 2009, Intel Corporation.
 *
 * Author: Raymond Liu <raymond.liu@intel.com>
 * 
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 * 
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */


#include <gtk/gtk.h>
#include "fvkbd.h"
#include "fvkbd-keyboard.h"
#include "misc-utility.h"
#include "gtk-misc-utility.h"

#include "fvkbd-key.h"

typedef struct _KeyUI KeyUI;
struct _KeyUI {
	GtkWidget *main_label;
	GtkWidget *sub_label;
	GtkWidget *img;

	PangoFontDescription *font_desc;
	PangoFontDescription *efont_desc;
	PangoFontDescription *ofont_desc;
};

typedef struct _CandidateItem CandidateItem;
struct _CandidateItem {
	GtkWidget *widget;
	KeyActionType type;
	gchar *disp;
	union 
	{
		gchar *string;
		KeySym sym;
		FvkbdKeyFunc func;
	} u;
};

#define MAX_CANDIDATE_ITEMS 10

static guint longpress_timeout_id = 0;
static guint longpress_timeout_val = 800;
static guint pop_window_hide_timeout_id = 0;
static guint pop_window_hide_timeout_val = 5000;

static GtkWidget *pop_window = NULL;
static CandidateItem candidate_items[MAX_CANDIDATE_ITEMS];

static gboolean key_mode_do_switch_lock = FALSE;


static void set_candidate_item_string(gint n, gchar *disp, gchar *str, PangoFontDescription *desc);
static void candidate_item_clicked_cb(GtkButton *button, CandidateItem *item);
static void init_candidate_items(void);
static void update_candidate_item_none(void);
static void update_candidate_item_string(gchar *disp, gchar *str, PangoFontDescription *desc, gboolean reverse);
static void update_candidate_item_string_group(gchar **strs, PangoFontDescription *desc, gboolean reverse);
static void update_candidate_item_sym(gchar *disp, KeySym sym, PangoFontDescription *desc, gboolean reverse);
static GtkWidget *get_pop_window(void);
static void settle_pop_window(FvkbdKey *key, gint number, gboolean reverse);
static void show_pop_window(FvkbdKey *key, gboolean longpress);
static void hide_pop_window(void);
static gboolean hide_pop_window_cb(FvkbdKey *key);
static gboolean longpress_detected_cb(FvkbdKey *key);
static PangoFontDescription * get_key_pango_font_description (FvkbdUnit *unit, KbdFontType type);


static void
set_candidate_item_dummy(gint n)
{
	GtkWidget *button;

	if (n >= MAX_CANDIDATE_ITEMS)
		return;

	candidate_items[n].type = KEY_ACTION_NONE;
	button = candidate_items[n].widget;
	gtk_button_set_label(GTK_BUTTON(button), "");
}


static void
set_candidate_item_string(gint n, gchar *disp, gchar *str, PangoFontDescription *desc)
{
	GtkWidget *button;
	GtkWidget *key_label;
	if (str == NULL || n >= MAX_CANDIDATE_ITEMS)
		return;

	candidate_items[n].type = KEY_ACTION_STRING;
	candidate_items[n].u.string = str;
	button = candidate_items[n].widget;
	gtk_button_set_label(GTK_BUTTON(button), disp);
	key_label = gtk_bin_get_child(GTK_BIN(button));
	gtk_widget_modify_font(key_label, desc);
}


static void
set_candidate_item_color(gint n, KbdColorType type, GdkColor *color)
{
	GtkWidget *button;
	GtkWidget *label;

	if (n >= MAX_CANDIDATE_ITEMS)
		return;

	if ((type != KBD_COLOR_TYPE_KEY_POP_BG) && (type != KBD_COLOR_TYPE_KEY_POP_FG))
		return;

	button = candidate_items[n].widget;
	if (type == KBD_COLOR_TYPE_KEY_POP_BG) {
		gtk_widget_modify_bg(button, GTK_STATE_NORMAL, color);
		gtk_widget_modify_bg(button, GTK_STATE_PRELIGHT, color);
		gtk_widget_modify_bg(button, GTK_STATE_ACTIVE, color);
	} else if (type == KBD_COLOR_TYPE_KEY_POP_FG){
		label = gtk_bin_get_child(GTK_BIN(button));
		if (label) {
			gtk_widget_modify_fg(label, GTK_STATE_NORMAL, color);
			gtk_widget_modify_fg(label, GTK_STATE_PRELIGHT, color);
			gtk_widget_modify_fg(label, GTK_STATE_ACTIVE, color);
		}
	}
}


static void
candidate_item_clicked_cb(GtkButton *button, CandidateItem *item)
{
	switch (item->type) {

	case KEY_ACTION_STRING:
		fvkbd_key_send_utf8_string(item->u.string);
		break;

	case KEY_ACTION_SYM:
		fvkbd_key_send_xkeysym(item->u.sym);
		break;

	default:
		break;
	}

	if (pop_window_hide_timeout_id != 0) {
		g_source_remove(pop_window_hide_timeout_id);
		pop_window_hide_timeout_id = 0;
		hide_pop_window();
	}

	if (fvkbd_in_temp_mode())
		fvkbd_keyboard_resume_default_mode();
}


static void
init_candidate_items(void)
{
	int i;
	for (i = 0; i < MAX_CANDIDATE_ITEMS; i++) {
		candidate_items[i].widget = gtk_button_new();
		gtk_button_set_relief(GTK_BUTTON(candidate_items[i].widget), GTK_RELIEF_NONE);
		g_signal_connect(G_OBJECT(candidate_items[i].widget), "clicked",
			G_CALLBACK(candidate_item_clicked_cb), &candidate_items[i]);
	}
}


static void
set_candidate_items_size(int width, int height)
{
	int i;
	for (i = 0; i < MAX_CANDIDATE_ITEMS; i++) {
		gtk_widget_set_size_request(candidate_items[i].widget,
						width, height);
	}
}


static void
update_candidate_item_none(void)
{
	int i;

	for (i = 0; i < MAX_CANDIDATE_ITEMS; i++)
		gtk_widget_hide(candidate_items[i].widget);
}


static void
update_candidate_item_string(gchar *disp, gchar *str, PangoFontDescription *desc,
				gboolean reverse)
{
	int i;

	if (!reverse) {
		set_candidate_item_dummy(0);
		set_candidate_item_string(1, disp, str, desc);
	} else {
		set_candidate_item_string(0, disp, str, desc);
		set_candidate_item_dummy(1);
	}

	gtk_widget_show(candidate_items[0].widget);
	gtk_widget_show(candidate_items[1].widget);

	for (i = 2; i < MAX_CANDIDATE_ITEMS; i++)
		gtk_widget_hide(candidate_items[i].widget);

}


static void
update_candidate_item_string_group(gchar **strs, PangoFontDescription *desc,
					gboolean reverse)
{
	int i;
	int n = 0;
	gchar **tmp;

	if (strs == NULL)
		return;

	for (i = 0, tmp = strs; (i < (MAX_CANDIDATE_ITEMS - 1)) && *tmp; i++, tmp++)
		n++;

	if (!reverse) {
		set_candidate_item_dummy(0);
		gtk_widget_show(candidate_items[0].widget);

		for (i = 1; i <= n; i++, strs++) {
			set_candidate_item_string(i, *strs, *strs, desc);
			gtk_widget_show(candidate_items[i].widget);
		}
	} else {
		for (i = 0; i < n; i++, strs++) {
			set_candidate_item_string(n - 1 - i, *strs, *strs, desc);
			gtk_widget_show(candidate_items[n - 1 - i].widget);
		}
		set_candidate_item_dummy(n);
		gtk_widget_show(candidate_items[n].widget);	

	}

	for (i = n + 1; i < MAX_CANDIDATE_ITEMS; i++)
		gtk_widget_hide(candidate_items[i].widget);
}


static void
update_candidate_item_sym(gchar *disp, KeySym sym, PangoFontDescription *desc,
				gboolean reverse)
{
	int i;
	GtkWidget *button;
	GtkWidget *key_label;

	if (!reverse) {
		set_candidate_item_dummy(0);
		i = 1;
	} else {
		set_candidate_item_dummy(1);
		i = 0;
	}

	gtk_widget_show(candidate_items[0].widget);
	gtk_widget_show(candidate_items[1].widget);

	candidate_items[i].type = KEY_ACTION_SYM;
	candidate_items[i].u.sym = sym;
	button = candidate_items[i].widget;
	gtk_button_set_label(GTK_BUTTON(button), disp);
	key_label = gtk_bin_get_child(GTK_BIN(button));
	gtk_widget_modify_font(key_label, desc);
	gtk_widget_show(button);

	for (i = 2; i < MAX_CANDIDATE_ITEMS; i++)
		gtk_widget_hide(candidate_items[i].widget);
}


static GtkWidget *
get_pop_window(void)
{
	if (pop_window == NULL) {
		GtkWidget *hbox, *frame;
		int i;

		pop_window = gtk_window_new(GTK_WINDOW_POPUP);
		gtk_window_set_accept_focus(GTK_WINDOW(pop_window), FALSE);
		gtk_window_set_resizable(GTK_WINDOW(pop_window), FALSE);

		frame = gtk_frame_new(NULL);
		gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_ETCHED_OUT);

		hbox = gtk_hbox_new(FALSE, 0);

		gtk_container_add (GTK_CONTAINER (frame), hbox);

		init_candidate_items();
		for (i = 0; i < MAX_CANDIDATE_ITEMS; i++) {
			gtk_box_pack_start(GTK_BOX(hbox), candidate_items[i].widget,
						FALSE, FALSE, 0);
		}

		gtk_widget_show_all(frame);

		gtk_container_add (GTK_CONTAINER (pop_window), frame);
	}

	return pop_window;
}


static void
settle_pop_window_color(FvkbdKey *key, gint number)
{
	GtkWidget *w = get_pop_window();
	GdkColor bgcolor, fgcolor;
	int i;

	get_gdkcolor(FVKBD_UNIT(key), KBD_COLOR_TYPE_KEY_POP_BG, &bgcolor);
	get_gdkcolor(FVKBD_UNIT(key), KBD_COLOR_TYPE_KEY_POP_FG, &fgcolor);

	gtk_widget_modify_bg(w, GTK_STATE_NORMAL, &bgcolor);

	for (i = 0; i < number; i++) {
		set_candidate_item_color(i, KBD_COLOR_TYPE_KEY_POP_BG, &bgcolor);
		set_candidate_item_color(i, KBD_COLOR_TYPE_KEY_POP_FG, &fgcolor);
	}
}


static void
settle_pop_window(FvkbdKey *key, gint number, gboolean reverse)
{
	GtkWidget *w = get_pop_window();
	GdkWindow *keywindow;

	int pop_w = 0, pop_h = 0;
	int x, y;
	int ux, uy, uw, uh;
	int ww, wh;
	gfloat x_ratio, y_ratio;
	GtkWidget *key_widget;

	key_widget = fvkbd_unit_get_widget(FVKBD_UNIT(key));
	keywindow = gtk_widget_get_window(key_widget);

	gdk_window_get_origin(keywindow, &x, &y);
	gtk_window_get_size(GTK_WINDOW(w), &pop_w, &pop_h);

	if ((fvkbd_unit_get_position(FVKBD_UNIT(key), &ux, &uy) == 0) &&
		(fvkbd_unit_get_size(FVKBD_UNIT(key), &uw, &uh) == 0) &&
		(fvkbd_unit_get_ratio(FVKBD_UNIT(key), &x_ratio, &y_ratio) == 0)) {
		if (!reverse) {
			x += (ux * x_ratio);
			y += (uy * y_ratio);
		} else {
			gtk_window_get_size(GTK_WINDOW(w), &ww, &wh);
			//x += ((ux - (uw * number)) * x_ratio);
			x += ((ux + uw) * x_ratio - ww);
			y += (uy * y_ratio);
		}
	}

	gtk_window_move(GTK_WINDOW(w), x, y);
}


static void
show_pop_window(FvkbdKey *key, gboolean longpress)
{
	FvkbdKeyAction *action;
	GtkWidget *w = get_pop_window();
	int item_number = 1;
	int i;
	int ux, uy, uw, uh, pw, ph;
	gfloat x_ratio, y_ratio;
	gboolean reverse = FALSE;

	fvkbd_unit_get_size(fvkbd_unit_get_parent(FVKBD_UNIT(key)), &pw, &ph);
	if ((fvkbd_unit_get_position(FVKBD_UNIT(key), &ux, &uy) == 0)) {
		if (ux <= (pw / 2))
			reverse = FALSE;
		else
			reverse = TRUE;
	}

	fvkbd_unit_get_ratio(FVKBD_UNIT(key), &x_ratio, &y_ratio);
	fvkbd_unit_get_size(FVKBD_UNIT(key), &uw, &uh);

	action = fvkbd_key_get_current_action(key);
	KeyUI *ui_data = fvkbd_unit_get_ui_data(FVKBD_UNIT(key));
	PangoFontDescription *desc;

	if (!ui_data->ofont_desc) {
		desc = get_key_pango_font_description(FVKBD_UNIT(key), KBD_FONT_TYPE_POP);
	} else
		desc = ui_data->ofont_desc;

	switch (action->type) {

	case KEY_ACTION_STRING:
		update_candidate_item_string(action->disp, action->u.string, desc, reverse);
		item_number = 2;
		break;

	case KEY_ACTION_STRING_GROUP:
		if (!longpress) {
			update_candidate_item_string(action->disp, action->u.string, desc, reverse);
			item_number = 2;
		} else {
			update_candidate_item_string_group(action->u.string_group, desc, reverse);
			for (i = 0; (i < MAX_CANDIDATE_ITEMS) && *(action->u.string_group + i); i++)
				item_number++;
		}
		break;

	case KEY_ACTION_SYM:
		update_candidate_item_sym(action->disp, action->u.sym, desc, reverse);
		item_number = 2;
		break;

	default:
		update_candidate_item_none();
		hide_pop_window();
		return;
	}

	set_candidate_items_size(uw * x_ratio, uh * y_ratio);
	settle_pop_window_color(key, item_number);
	// well, hiding pop window here, just wish the window size will be retrived
	// correctly by gtk_window_get_size later in settle_pop_window.
	hide_pop_window();
	settle_pop_window(key, item_number, reverse);
	if (!GTK_WIDGET_VISIBLE(w))
		gtk_widget_show(w);
}


static void
hide_pop_window(void)
{
	GtkWidget *w = get_pop_window();
	if (GTK_WIDGET_VISIBLE(w))
		gtk_widget_hide(w);
}


static gboolean
hide_pop_window_cb(FvkbdKey *key)
{
	hide_pop_window();
	pop_window_hide_timeout_id = 0;
	
	return FALSE;
}


static gboolean
longpress_detected_cb(FvkbdKey *key)
{
	FvkbdKeyAction *action;

	action = fvkbd_key_get_current_action(key);

	switch (action->type) {

	case KEY_ACTION_STRING:
	case KEY_ACTION_STRING_GROUP:
	case KEY_ACTION_SYM:
		show_pop_window(key, TRUE);
		pop_window_hide_timeout_id = g_timeout_add(pop_window_hide_timeout_val,
						(GSourceFunc)hide_pop_window_cb, key);

		break;
	case KEY_ACTION_FUNC:
		switch (action->u.func.type) {
		case FVKBD_KEY_FUNC_MODE_SELECT:
			key_mode_do_switch_lock = TRUE;
			break;

		default:
			break;
		}

		break;
	default:
		break;
	}

	longpress_timeout_id = 0;
	
	return FALSE;
}


static void
fvkbd_key_gtk_pressed_cb(GtkButton *button, FvkbdKey *key)
{
	if (longpress_timeout_id != 0) {
		g_source_remove(longpress_timeout_id);
		longpress_timeout_id = 0;
	}

	if (pop_window_hide_timeout_id != 0) {
		g_source_remove(pop_window_hide_timeout_id);
		pop_window_hide_timeout_id = 0;
		hide_pop_window();
	}

	longpress_timeout_id = g_timeout_add(longpress_timeout_val,
					(GSourceFunc)longpress_detected_cb, key);

	show_pop_window(key, FALSE);
}


static void
fvkbd_key_gtk_released_cb(GtkButton *button, FvkbdKey *key)
{
	FvkbdKeyAction *action;
	gboolean need_reset_default_mode = TRUE;

	/* if the candiate window is shown, then we do nothing and return */
	if (pop_window_hide_timeout_id != 0) {
		return;
	}

	if (longpress_timeout_id != 0) {
		g_source_remove(longpress_timeout_id);
		longpress_timeout_id = 0;
	}

	hide_pop_window();

	if ((action = fvkbd_key_get_current_action(key)) == NULL)
		return;

	switch (action->type) {
	case KEY_ACTION_STRING:
		fvkbd_key_send_utf8_string(action->u.string);
		break;

	case KEY_ACTION_STRING_GROUP:
		fvkbd_key_send_utf8_string(action->u.string_group[0]);
		break;

	case KEY_ACTION_SYM:
		fvkbd_key_send_xkeysym(action->u.sym);
		break;

	case KEY_ACTION_FUNC:
		switch (action->u.func.type) {
		case FVKBD_KEY_FUNC_MODE_SELECT:
			fvkbd_key_func_handlers(key, action, &key_mode_do_switch_lock);
			need_reset_default_mode = FALSE;
			key_mode_do_switch_lock = FALSE;
			break;

		default:
			fvkbd_key_func_handlers(key, action, NULL);
		}

		break;

	case KEY_ACTION_SCRIPT:
		fvkbd_do_script(action->u.string);
		break;

	default:
		break;
	}

	if (fvkbd_in_temp_mode() && need_reset_default_mode)
		fvkbd_keyboard_resume_default_mode();

}


static void
fvkbd_key_settle_color (FvkbdUnit *unit, GtkWidget *key_widget, gboolean lock)
{
	KeyUI *ui_data = fvkbd_unit_get_ui_data(unit);
	GdkColor color;
	GtkWidget *label;

	// set key background color
	if (!lock) {
		if (get_gdkcolor(unit, KBD_COLOR_TYPE_KEY_BG, &color)) {
			gtk_widget_modify_bg(key_widget, GTK_STATE_NORMAL, &color);
			gtk_widget_modify_bg(key_widget, GTK_STATE_PRELIGHT, &color);
		}

		if (get_gdkcolor(unit, KBD_COLOR_TYPE_KEY_FG, &color)) {
			gtk_widget_modify_bg(key_widget, GTK_STATE_ACTIVE, &color);
		}
	} else {
		if (get_gdkcolor(unit, KBD_COLOR_TYPE_KEY_FG, &color)) {
			gtk_widget_modify_bg(key_widget, GTK_STATE_NORMAL, &color);
			gtk_widget_modify_bg(key_widget, GTK_STATE_PRELIGHT, &color);
		}

		if (get_gdkcolor(unit, KBD_COLOR_TYPE_KEY_BG, &color)) {
			gtk_widget_modify_bg(key_widget, GTK_STATE_ACTIVE, &color);
		}
	}

	// set main label color
	label = ui_data->main_label;
	if (label) {
		if (!lock) {
			if (get_gdkcolor(unit, KBD_COLOR_TYPE_KEY_FG, &color)) {
				gtk_widget_modify_fg(label, GTK_STATE_NORMAL, &color);
				gtk_widget_modify_fg(label, GTK_STATE_PRELIGHT, &color);
			}

			if (get_gdkcolor(unit, KBD_COLOR_TYPE_KEY_BG, &color)) {
				gtk_widget_modify_fg(label, GTK_STATE_ACTIVE, &color);
			}
		} else {
			if (get_gdkcolor(unit, KBD_COLOR_TYPE_KEY_BG, &color)) {
				gtk_widget_modify_fg(label, GTK_STATE_NORMAL, &color);
				gtk_widget_modify_fg(label, GTK_STATE_PRELIGHT, &color);
			}

			if (get_gdkcolor(unit, KBD_COLOR_TYPE_KEY_FG, &color)) {
				gtk_widget_modify_fg(label, GTK_STATE_ACTIVE, &color);
			}
		}
	}

	// set sub label color
	label = ui_data->sub_label;
	if (label) {
		if (get_gdkcolor(unit, KBD_COLOR_TYPE_KEY_EXTRA_FG, &color)) {
			gtk_widget_modify_fg(label, GTK_STATE_NORMAL, &color);
			gtk_widget_modify_fg(label, GTK_STATE_PRELIGHT, &color);
		}

		if (get_gdkcolor(unit, KBD_COLOR_TYPE_KEY_BG, &color)) {
			gtk_widget_modify_fg(label, GTK_STATE_ACTIVE, &color);
		}
	}
}


static PangoFontDescription *
get_key_pango_font_description (FvkbdUnit *unit, KbdFontType type)
{
	gfloat x_ratio, y_ratio, ratio;
	int dpi_xres, dpi_yres;
	KeyUI *ui_data = fvkbd_unit_get_ui_data(unit);
	gchar *font_family = fvkbd_key_get_font_family(unit, type);
	gchar *font_weight = fvkbd_key_get_font_weight(unit, type);
	gint font_size = fvkbd_key_get_font_size(unit, type);

	PangoFontDescription *desc = NULL;

	fvkbd_unit_get_ratio(unit, &x_ratio, &y_ratio);
	get_resolution(&dpi_xres, &dpi_yres);

	x_ratio = x_ratio * DEFAULT_DPI / dpi_xres;
	y_ratio = y_ratio * DEFAULT_DPI / dpi_yres;

	ratio = (x_ratio < y_ratio) ? x_ratio : y_ratio;
	font_size = font_size * ratio;

	switch (type) {
		case KBD_FONT_TYPE_NORMAL:
			if (!ui_data->font_desc)
				ui_data->font_desc = get_pango_font_description_from_info(font_family, font_weight, font_size);
			desc = ui_data->font_desc;
			break;

		case KBD_FONT_TYPE_EXTRA:
			if (!ui_data->efont_desc)
				ui_data->efont_desc = get_pango_font_description_from_info(font_family, font_weight, font_size);
			desc = ui_data->efont_desc;
			break;

		case KBD_FONT_TYPE_POP:
			if (!ui_data->ofont_desc)
				ui_data->ofont_desc = get_pango_font_description_from_info(font_family, font_weight, font_size);
			desc = ui_data->ofont_desc;
			break;

		default:
			break;
	}

	return desc;
}


static void
fvkbd_key_settle_font (FvkbdUnit *unit, GtkWidget *key_widget)
{
	KeyUI *ui_data = fvkbd_unit_get_ui_data(unit);
	PangoFontDescription *desc;
	GtkWidget *label;

	// set main label font
	label = ui_data->main_label;
	if (label) {
		if (!ui_data->font_desc) {
			desc = get_key_pango_font_description(unit, KBD_FONT_TYPE_NORMAL);
		} else
			desc = ui_data->font_desc;

		gtk_widget_modify_font(label, desc);
	}

	// set sub label font
	label = ui_data->sub_label;
	if (label) {
		if (!ui_data->efont_desc) {
			desc = get_key_pango_font_description(unit, KBD_FONT_TYPE_EXTRA);
		} else
			desc = ui_data->efont_desc;

			gtk_widget_modify_font(label, desc);
	}
}


static GtkWidget *
_create_key_label (FvkbdUnit *unit, gint id)
{
	FvkbdKey *key;
	FvkbdKeyAction *action;
	KeyUI *ui_data = fvkbd_unit_get_ui_data(unit);
	gint w,h;
	GtkWidget *vbox = gtk_vbox_new(FALSE, 0);
	GtkWidget *label = NULL, *sub_label = NULL, *img = NULL;

	g_return_val_if_fail(FVKBD_IS_KEY(unit), FALSE);
	key = FVKBD_KEY(unit);

	fvkbd_unit_get_size(unit, &w, &h);

	if ((action = fvkbd_key_get_current_action(key)) == NULL)
		goto done;

	if (action->img != NULL) {
		img = gtk_image_new_from_file(action->img);
		goto done;
	}

	if (action->type == KEY_ACTION_STRING_GROUP) {
		gchar *label_str = NULL;

		#if 0
		// show all the optional characters
		label_str = g_strjoinv(" ", &(action->u.string_group[1]));
		#else
		// only show the first optional character
		label_str = g_strdup(action->u.string_group[1]);
		#endif

		sub_label = gtk_label_new(label_str);
		g_free(label_str);
		label = gtk_label_new(fvkbd_key_get_disp(key, id));
	} else if (action->type == KEY_ACTION_STRING ||
		action->type == KEY_ACTION_SYM ||
		action->type == KEY_ACTION_FUNC ||
		action->type == KEY_ACTION_SCRIPT) {

		sub_label = gtk_label_new(NULL);
		label = gtk_label_new(fvkbd_key_get_disp(key, id));
	}

done:

	ui_data->sub_label = sub_label;
	if (sub_label)
		gtk_box_pack_start(GTK_BOX(vbox), sub_label, FALSE, FALSE, 0);

	ui_data->main_label = label;
	if (label)
		gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

	ui_data->img = img;
	if (img)
		gtk_box_pack_start(GTK_BOX(vbox), img, TRUE, TRUE, 0);

	gtk_widget_show_all(vbox);
	return vbox;
}


static void
_update_key_label (FvkbdUnit *unit, gint id)
{
	GtkWidget *key_widget;
	GtkWidget *key_label;

	key_widget = fvkbd_unit_get_widget(unit);
	
	key_label = gtk_bin_get_child(GTK_BIN(key_widget));

	if (key_label) {
		gtk_widget_destroy(key_label);
		key_label = NULL;
	}

	key_label = _create_key_label(unit, id);
	gtk_container_add(GTK_CONTAINER(key_widget), key_label);
}


static void
_set_key_label_tmp_mode (FvkbdUnit *unit)
{
	KeyUI *ui_data = fvkbd_unit_get_ui_data(unit);
	GtkWidget *label;
	gchar *disp;

	label = ui_data->main_label;
	disp = g_strconcat("<u>", gtk_label_get_text(GTK_LABEL(label)), "</u>", NULL);
	gtk_label_set_text(GTK_LABEL(label), disp);
	gtk_label_set_use_markup(GTK_LABEL(label), TRUE);
	g_free(disp);
}


gboolean
fvkbd_key_build_ui (FvkbdUnit *unit, gpointer *widget)
{
	FvkbdKey *key;
	KeyUI *ui_data;
	gint w,h;
	gboolean ret = TRUE;

	GtkWidget *key_widget;
	GtkWidget *key_label;

	g_return_val_if_fail(FVKBD_IS_KEY(unit), FALSE);
	key = FVKBD_KEY(unit);

	key_widget = gtk_button_new();

	ui_data = g_new0(KeyUI, 1);
	fvkbd_unit_set_ui_data(unit, ui_data);
	ui_data->main_label = NULL;
	ui_data->sub_label = NULL;
	ui_data->font_desc = NULL;
	ui_data->efont_desc = NULL;
	ui_data->ofont_desc = NULL;

	fvkbd_unit_get_size(unit, &w, &h);
	gtk_widget_set_size_request(key_widget, w, h);

	key_label = _create_key_label(unit, 0);
	gtk_container_add(GTK_CONTAINER(key_widget), key_label);

	// set color
	fvkbd_key_settle_color(unit, key_widget, FALSE);

	// set font
	fvkbd_key_settle_font(unit, key_widget);

	g_signal_connect(G_OBJECT(key_widget), "pressed",
			G_CALLBACK(fvkbd_key_gtk_pressed_cb), key);
	g_signal_connect(G_OBJECT(key_widget), "released",
			G_CALLBACK(fvkbd_key_gtk_released_cb), key);

	if (widget != NULL)
		*widget = key_widget;
	fvkbd_unit_set_widget(unit, key_widget);
	return ret;
}


gboolean
fvkbd_key_resize_ui (FvkbdUnit *unit, gfloat x_ratio, gfloat y_ratio)
{
	FvkbdKey *key;
	KeyUI *ui_data;
	GtkWidget *key_widget;
	//GtkWidget *key_label;
	int tmp_w, tmp_h;
	int new_w, new_h;

	g_return_val_if_fail(FVKBD_IS_KEY(unit), FALSE);
	key = FVKBD_KEY(unit);

	ui_data = fvkbd_unit_get_ui_data(unit);
	key_widget = fvkbd_unit_get_widget(unit);

	if (fvkbd_unit_get_size(unit, &tmp_w, &tmp_h) == 0) {
		new_w = tmp_w * x_ratio;
		new_h = tmp_h * y_ratio;
		gtk_widget_set_size_request(key_widget, new_w, new_h);

	} else
		return FALSE;

	if (ui_data->font_desc) {
		pango_font_description_free(ui_data->font_desc);
		ui_data->font_desc = NULL;
	}
	if (ui_data->efont_desc) {
		pango_font_description_free(ui_data->efont_desc);
		ui_data->efont_desc = NULL;
	}
	if (ui_data->ofont_desc) {
		pango_font_description_free(ui_data->ofont_desc);
		ui_data->ofont_desc = NULL;
	}

	fvkbd_unit_set_ratio(unit, x_ratio, y_ratio);
	fvkbd_key_settle_font(unit, key_widget);

	return TRUE;
}


gboolean
fvkbd_key_destroy_ui (FvkbdUnit *unit)
{
	FvkbdKey *key;
	KeyUI *ui_data;
	gboolean ret = TRUE;

	GtkWidget *key_widget;

	g_return_val_if_fail(FVKBD_IS_KEY(unit), FALSE);
	key = FVKBD_KEY(unit);

	key_widget = fvkbd_unit_get_widget(unit);

	gtk_widget_hide_all(key_widget);
	gtk_widget_destroy(key_widget);

	fvkbd_unit_set_widget(unit, NULL);

	ui_data = fvkbd_unit_get_ui_data(unit);
	if (ui_data->font_desc)
		pango_font_description_free(ui_data->font_desc);
	if (ui_data->efont_desc)
		pango_font_description_free(ui_data->efont_desc);
	if (ui_data->ofont_desc)
		pango_font_description_free(ui_data->ofont_desc);
	g_free(ui_data);
	fvkbd_unit_set_ui_data(unit, NULL);

	return ret;
}


gint
fvkbd_key_set_mode (FvkbdUnit *unit, gint id)
{
	FvkbdKey *key;
	FvkbdKeyAction *action;
	GtkWidget *key_widget;
	gboolean show_lock_mode = FALSE;

	g_return_val_if_fail(FVKBD_IS_KEY(unit), -1);
	key = FVKBD_KEY(unit);

	key_widget = fvkbd_unit_get_widget(unit);

	// Yes, we get the default action here to handle disp diffirently for tmply mode change
	action = fvkbd_key_get_action(key, 0);

	if ((action->type == KEY_ACTION_FUNC) &&
		(action->u.func.type == FVKBD_KEY_FUNC_MODE_SELECT)) {
		if ((fvkbd_in_temp_mode()) && ((*((int *)action->u.func.data)) == id)) {
			_update_key_label(unit, 0);
			_set_key_label_tmp_mode(unit);
		} else {
			_update_key_label(unit, id);
			if ((*((int *)action->u.func.data)) == id)
				show_lock_mode = TRUE;
		}
	} else {
		_update_key_label(unit, id);
	}

	// set color
	fvkbd_key_settle_color(unit, key_widget, show_lock_mode);

	// set font
	fvkbd_key_settle_font(unit, key_widget);

	return 0;
}

