/*
 * fvkbd-panel-ui-gtk.c panel unit gtk ui for fvkbd  
 *
 * Copyright (C) 2009, Intel Corporation.
 *
 * Author: Raymond Liu <raymond.liu@intel.com>
 * 
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 * 
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#include <gtk/gtk.h>
#include "fvkbd.h"
#include "gtk-misc-utility.h"

#include "fvkbd-panel.h"

typedef struct _PanelUI PanelUI;
struct _PanelUI {
	int alloc_w;
	int alloc_h;
};


static void
panel_container_on_realize (GtkWidget *widget, gpointer data)
{
	FvkbdUnit *unit = FVKBD_UNIT(data);
	FvkbdPanel *panel = FVKBD_PANEL(unit);

	GtkWidget *panel_widget;
	gchar *img = NULL;
	KbdColor *kc;
	guint color = 0xFFFFFFFF;

	if ((img = fvkbd_panel_get_img(panel)) == NULL)
		return;

	if ((kc = fvkbd_unit_get_color(unit, KBD_COLOR_TYPE_PANEL_BG)) != NULL)
		color = (kc->r << 24) + (kc->g << 16) + (kc->b << 8) + (kc->a);

	panel_widget = fvkbd_unit_get_widget(unit);

	set_gtk_widget_bg_image(panel_widget, img, color);

	return;
}


static void
panel_container_size_allocate (GtkWidget *widget, GtkAllocation *allocation, gpointer data)
{
	FvkbdUnit *unit = FVKBD_UNIT(data);
	PanelUI *ui_data;
	int tmp_w = 1, tmp_h = 1;
	float x_ratio, y_ratio;

	ui_data = fvkbd_unit_get_ui_data(unit);

	// don't do child placement when allocation not change, or it will cause this call back been invoked looply.
	if ((ui_data->alloc_w == allocation->width) && (ui_data->alloc_h == allocation->height))
		return;
	ui_data->alloc_w = allocation->width;
	ui_data->alloc_h = allocation->height;

	fvkbd_unit_get_size(unit, &tmp_w, &tmp_h);
	x_ratio = (float)allocation->width / (float)tmp_w;
	y_ratio = (float)allocation->height / (float)tmp_h;

	fvkbd_unit_resize_ui(unit, x_ratio, y_ratio);

	return;
}


gboolean
fvkbd_panel_build_ui (FvkbdUnit *unit, gpointer *widget)
{
	FvkbdPanel *panel;
	GSList *children;
	FvkbdUnit *child;
	gboolean ret = TRUE;
	PanelUI *ui_data;
	GtkWidget *child_widget = NULL;
	GtkWidget *fixed_widget = NULL;
	GdkColor bgcolor;
	gint x = 0, y = 0;
	gint i;

	g_return_val_if_fail(FVKBD_IS_PANEL(unit), FALSE);
	panel = FVKBD_PANEL(unit);

	ui_data = g_new0(PanelUI, 1);
	ui_data->alloc_w = 0;
	ui_data->alloc_h = 0;

	fixed_widget = gtk_fixed_new();
	gtk_fixed_set_has_window(GTK_FIXED(fixed_widget), TRUE);

	if (get_gdkcolor(unit, KBD_COLOR_TYPE_PANEL_BG, &bgcolor)) {
		gtk_widget_modify_bg(fixed_widget, GTK_STATE_NORMAL, &bgcolor);
	}

	g_signal_connect(fixed_widget, "size-allocate",
			G_CALLBACK(panel_container_size_allocate), unit);

	g_signal_connect(fixed_widget, "realize",
			G_CALLBACK(panel_container_on_realize), unit);


	// packing child widget
	children = fvkbd_panel_get_children(panel);
	if (children == NULL)
		goto done;

	for (i = 0; i < g_slist_length(children); i++) {
		child = g_slist_nth_data(children, i);
		if ((fvkbd_get_unit_type(child) != UNIT_TYPE_KEY)) {
			continue;
		}

		if ((ret  = fvkbd_unit_build_ui(child, (gpointer *)&child_widget)) != TRUE)
			goto done;

		if (fvkbd_unit_get_position(child, &x, &y) == 0) {
			gtk_fixed_put(GTK_FIXED(fixed_widget), child_widget, x, y);
		}
	}

	if (widget != NULL)
		*widget = fixed_widget;
	fvkbd_unit_set_ui_data(unit, ui_data);
	fvkbd_unit_set_widget(unit, fixed_widget);
	gtk_widget_show_all(fixed_widget);
done:
	return ret;
}


gboolean
fvkbd_panel_resize_ui (FvkbdUnit *unit, gfloat x_ratio, gfloat y_ratio)
{
	FvkbdPanel *panel = FVKBD_PANEL(unit);
	GtkWidget *panel_widget;
	GSList *children;
	FvkbdUnit *child;
	GtkWidget *child_widget;
	GtkWidget *child_container;
	int base_x, base_y, new_x, new_y;
	int tmp_w = 1, tmp_h = 1;
	int new_w, new_h;
	int i;

	fvkbd_unit_set_ratio(unit, x_ratio, y_ratio);
	if (fvkbd_unit_get_size(unit, &tmp_w, &tmp_h) == 0) {
		panel_widget = fvkbd_unit_get_widget(unit);
		new_w = tmp_w * x_ratio;
		new_h = tmp_h * y_ratio;
		gtk_widget_set_size_request(panel_widget, new_w, new_h);
	} else
		return FALSE;

	child_container = fvkbd_unit_get_widget(unit);

	children = fvkbd_panel_get_children(panel);
	for (i = 0; i < g_slist_length(children); i++) {
		child = g_slist_nth_data(children, i);
		child_widget = fvkbd_unit_get_widget(child);
		if ((fvkbd_unit_get_position(child, &base_x, &base_y) == 0) &&
			(fvkbd_unit_get_size(child, &tmp_w, &tmp_h) == 0)) {
			new_x = base_x * x_ratio;
			new_y = base_y * y_ratio;

			/* FIXME: This is a tmp solution, it will make window shrink difficult */
			/* need to find another way to do child replacement and resize */
			/* maybe the right way to do it is to inherit from gtkfixed and override size_allocate function */
			gtk_fixed_move(GTK_FIXED(child_container), child_widget, new_x, new_y);
			fvkbd_unit_resize_ui(child, x_ratio, y_ratio);
		}
	}
	return TRUE;
}


gboolean
fvkbd_panel_destroy_ui (FvkbdUnit *unit)
{
	FvkbdPanel *panel;
	GSList *children;
	FvkbdUnit *child;
	gboolean ret = TRUE;
	PanelUI *ui_data;
	GtkWidget *panel_widget;
	gint i;

	g_return_val_if_fail(FVKBD_IS_PANEL(unit), FALSE);
	panel = FVKBD_PANEL(unit);

	ui_data = fvkbd_unit_get_ui_data(unit);
	panel_widget = fvkbd_unit_get_widget(unit);

	gtk_widget_hide_all(panel_widget);

	// destroy child widget
	children = fvkbd_panel_get_children(panel);
	if (children == NULL)
		goto done;

	for (i = 0; i < g_slist_length(children); i++) {
		child = g_slist_nth_data(children, i);
		if ((fvkbd_get_unit_type(child) != UNIT_TYPE_KEY)) {
			continue;
		}

		if ((fvkbd_unit_destroy_ui(child)) != TRUE)
			ret = FALSE;
	}

	gtk_widget_destroy(panel_widget);
	fvkbd_unit_set_widget(unit, NULL);
	fvkbd_unit_set_ui_data(unit, NULL);
	g_free(ui_data);
done:
	return ret;
}

